//
//  VisionViewController.swift
//  MachineLearningSample
//
//  Created by Toshihiro Morimoto on 2017/10/01.
//  Copyright © 2017 Toshihiro Morimoto. All rights reserved.
//

import UIKit
import Vision
import AVFoundation

class VisionViewController: UIViewController {
    @IBOutlet weak var imageView: UIImageView!
    @IBOutlet weak var visualizeBoxView: VisualizeBoxView!
    
    var session: AVCaptureSession?
    
    override func viewDidAppear(_ animated: Bool) {
        super.viewDidAppear(animated)
        start()
    }
    
    override func viewWillDisappear(_ animated: Bool) {
        stop()
        super.viewWillDisappear(animated)
    }
    
    private func start() {
        stop()
        session = AVCaptureSession()
        session?.sessionPreset = .photo
        
        let device = AVCaptureDevice.default(for: .video)!
        let input = try! AVCaptureDeviceInput(device: device)
        session?.addInput(input)
        
        let output = AVCaptureVideoDataOutput()
        output.videoSettings = [
            kCVPixelBufferPixelFormatTypeKey as String: kCVPixelFormatType_32BGRA
        ]
        output.setSampleBufferDelegate(self, queue: DispatchQueue.global())
        session?.addOutput(output)
        
        session?.startRunning()
    }
    
    private func stop() {
        session?.stopRunning()
    }
}

extension VisionViewController: AVCaptureVideoDataOutputSampleBufferDelegate {
    // カメラ入力を処理
    func captureOutput(_ output: AVCaptureOutput, didOutput sampleBuffer: CMSampleBuffer, from connection: AVCaptureConnection) {
        let image = self.image(from: sampleBuffer)
        let orientation = convertOrientation(from: image.imageOrientation)
        
        DispatchQueue.main.async { [weak self] in
            self?.imageView.image = image
        }
        
        // Vision によるテキスト検出
        let request = VNDetectTextRectanglesRequest() { request, error in
            guard let results = request.results as? [VNTextObservation] else { return }
            
            // テキストブロックの矩形情報
            let boundingBoxes = results.map { $0.boundingBox }
            // 文字の矩形情報
            let characterBoxes = results.flatMap {
                ($0.characterBoxes ?? []).map { $0.boundingBox }
            }
            
            // 検出した矩形情報をレンダリング
            DispatchQueue.main.async { [weak self] in
                self?.visualizeBoxView.boundingBoxes = boundingBoxes
                self?.visualizeBoxView.characterBoxes = characterBoxes
                self?.visualizeBoxView.setNeedsDisplay()
            }
        }
        request.reportCharacterBoxes = true
        
        let ciImage = CIImage(image: image)!
        let handler = VNImageRequestHandler(ciImage: ciImage, orientation: orientation)
        try! handler.perform([request])
    }
    
    private func image(from sampleBuffer: CMSampleBuffer) -> UIImage {
        let imageBuffer = CMSampleBufferGetImageBuffer(sampleBuffer)!
        
        CVPixelBufferLockBaseAddress(imageBuffer, .readOnly)
        defer {
            CVPixelBufferUnlockBaseAddress(imageBuffer, .readOnly)
        }
        
        let base = CVPixelBufferGetBaseAddress(imageBuffer)
        let bytesPerRow = CVPixelBufferGetBytesPerRow(imageBuffer)
        let width = CVPixelBufferGetWidth(imageBuffer)
        let height = CVPixelBufferGetHeight(imageBuffer)
        
        let colorSpace = CGColorSpaceCreateDeviceRGB()
        let bitmapInfo = CGBitmapInfo.byteOrder32Little.rawValue | CGImageAlphaInfo.premultipliedFirst.rawValue
        let context = CGContext(
            data: base,
            width: width,
            height: height,
            bitsPerComponent: 8,
            bytesPerRow: bytesPerRow,
            space: colorSpace,
            bitmapInfo: bitmapInfo)!
        
        let cgImage = context.makeImage()!
        return UIImage(cgImage: cgImage, scale: 1.0, orientation: .right)
    }
    
    private func convertOrientation(from orientation: UIImageOrientation) -> CGImagePropertyOrientation {
        let orientations: [CGImagePropertyOrientation] = [
            .up, .down, .left, .right, .upMirrored, .downMirrored, .leftMirrored, .rightMirrored
        ]
        return orientations[orientation.rawValue]
    }
}

class VisualizeBoxView: UIView {
    var boundingBoxes: [CGRect] = []
    var characterBoxes: [CGRect] = []
    
    override func draw(_ rect: CGRect) {
        backgroundColor = UIColor.clear
        
        func drawRect(_ rect: CGRect) {
            let path = UIBezierPath(rect: convertRect(rect: rect, to: frame.size))
            path.lineWidth = 2
            path.stroke()
        }
        
        UIColor.green.withAlphaComponent(0.8).setStroke()
        boundingBoxes.forEach {
            drawRect($0)
        }

        UIColor.yellow.withAlphaComponent(0.8).setStroke()
        characterBoxes.forEach {
            drawRect($0)
        }
    }
    
    private func convertRect(rect: CGRect, to size: CGSize) -> CGRect {
        // UIView の座標系に変換
        return CGRect(
            x: rect.minX * size.width,
            y: (1.0 - rect.maxY) * size.height,
            width: rect.width * size.width,
            height: rect.height * size.height)
    }
}
