//: 3章で利用するSwiftコードを集めたPlaygroundです。

import Foundation

/*:
 ## [SE-0163] String型の再評価
 */

/*:
 ### Collectionに再適合したString型
 */
do {
    let string = "WEB+DB PRESS"
    
    // for文によるイテレーション
    for char in string {
        print(char)
    }
    print("---")
}

/*:
 ### Substring型
 */
do {
    let original = "WEB+DB PRESS"
    let newString: String
    #if swift(>=4)
        let index = original.index(of: "+")!
        let substring = original[index..<original.endIndex]
        type(of: substring)
        // +DB PRESSがコピーされ、新たなバッファが確保される
        newString = String(substring)
        
    #else // <= swift 3.2
        let index = original.characters.index(of: "+")!
        // +DB PRESSではあるが、original変数全体のバッファへ参照を持つ
        newString = original[index..<original.endIndex]
        
        type(of: newString)
    #endif
    newString
}


/*:
 ### Unicode 9への対応
 */
"👨‍👩‍👧‍👦".count


/*:
 ## [SE-0168] 複数行の文字列リテラル
 */

do {
    // 複数行文字列リテラルの宣言
    let multilineString = """
        複数行の文字列リテラルに
        対応しました。
        """
    multilineString
    
    // 複数行に及ぶJSON文字列の宣言
    let JSONString = """
        {
            "id": "abc-123-def",
            "name": "Yuki Takei",
            "age": 29
        }
        """
    JSONString
}

/*:
 ## [SE-0165] Dictionary<Key, Value>型とSet<Element>型の拡張
 */

/*:
 ### SequenceからDictionary<Key, Value>型の初期化と結合
 */
do {
    let tuples = [("a", 1), ("b", 2), ("c", 3), ("d", 4)]
    let dict1 = Dictionary(uniqueKeysWithValues: tuples)
    dict1

    // Sequenceの重複を除いたディクショナリを生成する
    let duplicates = [("a", 1), ("b", 2), ("a", 3)]

    // uniquingKeysWithにmin関数を渡すことで、重複キーにおける最小値を選択する
    let uniques = Dictionary(duplicates, uniquingKeysWith: min)
    uniques
}

// mergeを使ったディクショナリ同士の結合
do {
    var dict1 = ["a": 1, "b": 2, "c": 3]
    let dict2 = ["a": 4, "b": 5]
    // keyが重複した場合に、min関数により最小値を選択する
    dict1.merge(dict2, uniquingKeysWith: min)
    dict1

    // mergingを使ったディクショナリとSequenceの結合
    let dict3 = ["a": 1, "b": 2, "c": 3]
    let tuples = [("a", 4), ("b", 5), ("c", 6), ("d", 7)]
    // keyが重複した場合に、max関数により最大値を選択する
    let merged = dict3.merging(tuples, uniquingKeysWith: max)
    merged
}


/*:
 ### Dictionary<Key, Value>型のサブスクリプトに対するデフォルト値
 */
do {
    var dictionary = ["a": 1]
    let n = dictionary["a", default: 0] + 1
    n
    
    // 即時的なmutatingメソッドを用いた結果はそのまま書き戻すことができる
    dictionary["b", default: 0] += 10
    dictionary
}


/*:
 ### Dictionary<Key, Value>型とSet<Element>型を返すfilterとmap
 */

// Dictionary
do {
    let numbers = ["one": 1, "two": 2, "three": 3]
    let numbers2 = numbers.mapValues({ $0 * 2 })
    type(of: numbers2)
    print(numbers2)
    print("---")
    
    let evens = numbers.filter { $0.value % 2 == 0 }
    type(of: evens)
    print(evens)
    print("---")
}

// Set
do {
    let numbers: Set<Int> = [1, 2, 3]
    let evens = numbers.filter { $0 % 2 == 0 }
    type(of: evens)
    print(evens)
    print("---")
}

/*:
 ### Sequence要素のグループ化
 */
do {
    // 要素の1文字目をkeyとしてグループ化されたディクショナリを生成
    let fruits = ["Apple", "Banana", "Apple", "Peach"]
    // byに渡すクロージャでは、Hashableな値を返すこと
    let grouped = Dictionary(grouping: fruits, by: { $0.first! })
    print(grouped)
    print("---")
}


/*:
 ## [SE-0104] プロトコル指向な整数型
 */
do {
    var x: Int8 = 42
    let y = 1
    let z = 0
    
    // Int8とIntによるビット演算
    x <<= y
    
    // Int8とIntによる比較
    x > z
}


/*:
 ## [SE-0172] 片側指定のみの範囲型
 */
do {
    let alphabets = ["a","b","c","d"]
    //lettersのcountまでを埋めたSequenceが生成される
    let numberedAlphabets = zip(1..., alphabets)
    print(Array(numberedAlphabets))
    print("---")
    
    let string = "WEB+DB PRESS"
    let index = string.index(of: "+")!
    string[index...]
}


/*:
 ## [SE-0142] 関連型を制約するwhere節の許可
 */

/*:
 ### 関連型にwhere節を指定
 */
protocol P1 {
    // Swift 3では、T.Iterator.Elementとしていたものを、
    // Swift 4では`T.Element`とショートカットできる
    associatedtype T1: Sequence where T1.Element == Int
}

/*:
 ### プロトコル宣言時のwhere節の指定
 */
protocol IntSequence : Sequence where Element == Int {}

struct Countdown: IntSequence, IteratorProtocol {
    var count: Int
    
    mutating func next() -> Int? {
        if count == 0 {
            return nil
        } else {
            defer { count -= 1 }
            return count
        }
    }
}

for n in Countdown(count: 2) {
    print(n)
}
print("---")

/*:
 ## [SE-0148] ジェネリックなサブスクリプト
 */

struct JSON {
    fileprivate var storage: [String: Any]
    
    init(dictionary: [String: Any]) {
        self.storage = dictionary
    }
    
    subscript<T>(key: String) -> T? {
        return storage[key] as? T
    }
}

do {
    let json = JSON(dictionary: [
        "name": "WEB+DB PRESS",
        "category": "Technology",
        "publisher": "技術評論社",
        "price": 1480
    ])
    
    let name: String? = json["name"]
    name
    let price: Int? = json["price"]
    price
}


/*:
 ## [SE-0169] private修飾子のスコープの変更
 */

struct S1 {
    private let value = 1
}

extension S1: CustomStringConvertible {
    var description: String {
        return "値: \(value)" // valueにアクセス可能
    }
}

print(S1())
print("---")


/*:
 ## objc属性の推論の制限
 */

// dynamicなメンバーに対するobjc属性の必須化
class MyClass {
    // error: 'dynamic' method must be '@objc'
    // dynamic func foo() { } // コメントアウトを外すとエラー
    
    // コンパイルできる
    @objc dynamic func bar() { }
}

// NSObjectを継承したクラスのメンバーに対するobjc属性の推論の廃止
class MyNSObject: NSObject {
    // Objective-Cから見えず、Swiftからのみ扱える
    func foo() { }
    
    // Objective-Cから見える
    @objc func bar() {}
}

/*:
 ### @objcMembersを使った推論の再有効化
 */
@objcMembers
class MyClass2 {
    dynamic func bar() { } // @objcが自動で推論される
}


/*:
 ## [SE-0166] Swiftらしいアーカイブとシリアライゼーション
 
 以下は、Swiftらしいアーカイブとシリアライゼーション節のフルセットのコードです。
*/

struct Tag: Codable {
    let name: String
}

enum Category: Int, Codable {
    case economy = 1
    case technology = 2
}

struct Article: Codable {
    let id: Int
    let title: String
    let body: String
    let postedAt: String?
    let tags: [Tag]
    let category: Category
    
    private enum CodingKeys: String, CodingKey {
        case id
        case title
        case body
        case postedAt = "posted_at"
        case tags
        case category
    }
}

do {
    let jsonString = """
    {
        "id": 1,
        "title": "タイトル",
        "body": "本文です。",
        "posted_at": "2017-01-01: 00:00:00",
        "tags": [
            {"name": "Swift 4"}
        ],
        "category": 2
    }
    """
    
    // JSON文字列からstructへのデコード
    let article = try JSONDecoder().decode(
        Article.self,
        from: jsonString.data(using: .utf8)!
    )
    
    print(article.id)
    print(article.postedAt!)
    print(article.tags[0].name)
    print(article.category)
    print("---")
    
    // structからJSON文字列へのエンコード
    let jsonData = try JSONEncoder().encode(article)
    print(String(data: jsonData, encoding: .utf8)!)
    print("---")
}
